from functools import wraps
from blues_lib.type.output.STDOut import STDOut
from blues_lib.deco.LogDeco import LogDeco

class BehaviorSTDOutLog(LogDeco):

  def __init__(self,title:str='behavior',max_chars:int=50):
    super().__init__(title,max_chars)

  def __call__(self,func):
    @wraps(func) 
    def wrapper(instance,*args,**kwargs):
      module_name:str = instance.__class__.__module__  # 模块路径（如 my_project.tasks.loop）
      self._log_config(module_name,instance)

      prefix:str = self._get_prefix(instance)
      stdout:STDOut = func(instance,*args,**kwargs)
      self._log(prefix,stdout)
      return stdout
    return wrapper
  
  def _get_prefix(self,instance)->str:
    module_name:str = instance.__class__.__module__  # 模块路径（如 my_project.tasks.loop）
    return f'{module_name} [{self._title}]'

  def _log_config(self,module_name:str,instance):
    ellipsis_config:dict = self._get_ellipsis_config(instance._config)
    self._logger.info(f"\n\n{module_name} config :: {ellipsis_config}")
    
  def _get_ellipsis_config(self,config:dict)->dict:
    ellipsis_config:dict = {}
    for key,value in config.items():
      if key == 'value' and self._max_chars>0 and len(str(value)) > self._max_chars:
        ellipsis_config[key] = str(value)[:self._max_chars]+' ...'
      else:
        ellipsis_config[key] = value
    return ellipsis_config


from blues_lib.util.ScriptExecutor import ScriptExecutor

class Python():
  
  def __init__(self,driver):
    self.driver = driver
  
  def execute(self,script:str):
    return ScriptExecutor.execute(script,self.driver)


from blues_lib.cleaner.CleanerHandler import CleanerHandler
from .MaterialHandler import MaterialHandler  
from .LogHandler import LogHandler  

class FileCleanerChain(CleanerHandler):

  kind = 'chain'

  def resolve(self,request):
    handler = self._get_chain()
    return handler.handle(request)

  def _get_chain(self):
    material_handler = MaterialHandler()
    log_handler = LogHandler()
    
    material_handler.set_next(log_handler)

    return material_handler



from typing import Type
from abc import abstractmethod
from blues_lib.type.executor.Executor import Executor
from .ProcFactory import ProcFactory
from .FuncHandler import FuncHandler

class Hook(Executor):

  def __init__(self,hook_defs:list[dict[str,any]],options:dict) -> None:
    '''
    @param {list[dict]} hook_defs : the task hook definition
    @param {dict} options: 可选参数，支持：
        - 其他未来可能扩展的参数（如 context、config 等）
    '''
    super().__init__()
    self._hook_defs = hook_defs
    self._options = options # 可变的扩展参数
    
  def _run(self,hook_def:dict)->any:
    kind:str = hook_def.get('kind') or ''
    value:str = hook_def.get('value') or ''
    if not kind or not value:
      return

    if kind == 'class':
      self._handle_by_class(hook_def)
    elif kind == 'func':
      self._handle_by_func(hook_def)
    else:
      raise ValueError(f'Unknown hook kind: {kind}')
    
  def _handle_by_class(self,hook_def:dict)->any:
    factory = self._get_proc_factory()
    if processor := factory.create(hook_def,self._options):
      processor.execute()
    else:
      raise Exception(f'processor {hook_def.get("value")} not found')
    
  def _handle_by_func(self,hook_def:dict)->any:
    handler = self._get_func_handler()
    handler(hook_def,self._options).execute()
  
  @abstractmethod
  def _get_proc_factory(self)->Type[ProcFactory]:
    pass
  
  @abstractmethod
  def _get_func_handler(self)->Type[FuncHandler]:
    pass


import os,base64
from datetime import datetime,timedelta
from blues_lib.util.OSystem import OSystem

class BluesFiler:
  
  @classmethod
  def get_lib_path(cls):
    # 此目录 用来存放此库生成的文件
    root_dir = "blues_lib"
    os_type = OSystem.get_os_type()
    if os_type == 'windows':
      return os.path.join("c:\\",root_dir)
    elif os_type == 'linux':
      return os.path.join('/usr/share',root_dir)
    elif os_type == 'mac':
      return os.path.join('/Users/Shared',root_dir)
    else:
      return ''

  @classmethod
  def get_cft_path(cls):
    # 此目录 存放chrome cft资源
    root_path = BluesFiler.get_lib_path()
    root_dir = 'cft'
    return os.path.join(root_path,root_dir)

  @classmethod
  def get_app_path(cls,project_root_dir:str='')->str:
    # 执行py文件的绝对路径
    cur_path = os.path.realpath(__file__)
    root_dir = project_root_dir
    return os.path.join(cur_path.split(root_dir)[0],root_dir)
  
  @classmethod
  def get_abs_path(cls,project_root_dir:str,path:str)->str:
    root_path = cls.get_app_path(project_root_dir)
    return os.path.join(root_path,path)

  @classmethod
  def readfiles(cls,directory):
    '''
    Read the file list in a dir, don't support the next dirs
    @param {string} directory 
    '''
    file_list = []
    for root, dirs, files in os.walk(directory):
      for file in files:
        file_list.append(os.path.join(root, file))
    return file_list

  @classmethod
  def removedirs(cls,directory,retention_days=0):
    '''
    @description Remove all child dir and files
    @param {string} directory 
    '''
    threshold = datetime.now() - timedelta(days=retention_days)
    removed_count = 0
    if not cls.exists(directory):
      return removed_count

    for root, dirs, files in os.walk(directory):
      for file in files:
        file_path = os.path.join(root, file)
        file_modified_time = datetime.fromtimestamp(os.path.getmtime(file_path))
        if file_modified_time < threshold:
          os.remove(file_path)
          removed_count +=1
      for dire in dirs:
        dir_path = os.path.join(root, dire)
        file_modified_time = datetime.fromtimestamp(os.path.getmtime(dir_path))
        if file_modified_time < threshold:
          # recursion to remove the dir
          removed_count += cls.removedirs(dir_path,retention_days)
    # remove the base dir
    if cls.is_dir_empty(directory):
      os.rmdir(directory)
      removed_count +=1
    return removed_count

  @classmethod
  def is_dir_empty(cls,dir_path):
    return not bool(os.listdir(dir_path))

  @classmethod
  def exists(cls,path):
    '''
    @description : Does a dir or file exist
    @param {str} path
    @returns {bool} 
    '''
    return os.path.exists(path)

  @classmethod
  def filter_exists(cls,files):
    exists_files = []
    for file in files:
      if not cls.exists(file):
        continue
      exists_files.append(file)
    return exists_files

  @classmethod
  def makedirs(cls,path):
    '''
    @description : Create dirs (support multilevel directory) if they don't exist
    @param {str} path : multilevel dir
    @returns {None}
    '''
    if not cls.exists(path):
      os.makedirs(path)

  @classmethod
  def get_rename_file(cls,file_path,new_name='',prefix='',suffix='',separator='-'):
    '''
    @description : get the new file name path
    '''
    path_slices = file_path.split('/')
    original_name = path_slices[-1]
    copy_name = new_name if new_name else original_name
    if prefix:
      copy_name = prefix+separator+copy_name
    if suffix:
      copy_name = copy_name+separator+suffix
    path_slices[-1]=copy_name
    copy_path='/'.join(path_slices)
    return copy_path

  @classmethod
  def removefiles(cls,directory,retention_days=0):
    '''
    @description : clear files before n days
    @param {str} directory
    @param {int} retention_days : default 7
    @returns {int} deleted files count
    '''
    # 转换天数到时间间隔
    threshold = datetime.now() - timedelta(days=retention_days)
    removed_count = 0
    # 遍历目录
    for item in os.scandir(directory):
      try:
        # 获取文件的最后修改时间
        file_modified_time = datetime.fromtimestamp(os.path.getmtime(item.path))
        # 如果文件的最后修改时间早于阈值，则删除文件
        if os.path.isfile(item.path) and file_modified_time < threshold:
          os.remove(item.path)
          removed_count +=1
      except OSError as e:
        pass

    return removed_count

  @classmethod
  def dump_base64(cls,text):
    # 只接受bytes类型 b'xx'，且返回bytes，将其转为字符串
    return base64.b64encode(text.encode()).decode()

  @classmethod
  def load_base64(cls,b64):
    # 只接受bytes类型 b'xx'，且返回bytes，将其转为字符串
    return base64.b64decode(b64.encode()).decode()



import re
from typing import Any
from pyhocon import ConfigFactory
from pyhocon.config_tree import ConfigTree
from blues_lib.hocon.replacer.Replacer import Replacer

class VariableFormatReplacer(Replacer):

  '''
  replace the variable in the template with the given values
  - single variable "{}"
  - interpolate variable "name is {}, age is {}"
  '''  

  VAR_SEARCH_PATTERN = r"\{\}"
  VAR_MATCH_PATTERN = r"^\{\}$"
  
  def replace(self)->dict:
    if not self._variables or not self._template:
      return self._template
    
    tree:ConfigTree = ConfigFactory.from_dict(self._template)
    
    for path,value in self._variables.items():
      node:dict = self._get_node(tree,path)
      node_key:str|int|None = node.get('key')
      node_value:Any|None = node.get('value')
      node_parent:ConfigTree|None = node.get('parent')
      
      # if the template's node value is not a string, skip (must not be a placeholder)
      if not isinstance(node_value,str):
        continue
      
      # if the node's value is not a placeholder, skip
      if not re.search(self.VAR_SEARCH_PATTERN,node_value):
        continue
      
      replacement:Any = None
      if re.match(self.VAR_MATCH_PATTERN,node_value):
        # if the placeholder is a single variable, replace it using the original value
        replacement = value
      else:
        # if the placeholder is an interpolate variable, replace it using the format syntax
        args:list[Any] = value if isinstance(value,list) else [value]
        replacement = node_value.format(*args)
      
      node_parent[node_key] = replacement
    
    return tree.as_plain_ordered_dict()
  
  def _get_node(self,tree:ConfigTree,path:str)->dict:
    '''
    get the ConfigTree's node info, support the path include list index:
    - with dot notation: "prompt.user.content"
    - with list index: "crawler.setup[0].urls[1].url"
    
    node's attributes:
    - key: the dict node's key or the list node's index
    - value: the node value (may be a placeholder)
    - parent: the parent node of the ConfigTree
    
    Error handling:
    - If the path doesn't exist, node's parent and value will be set to None
    - If array index is out of range or negative, node's parent and value will be set to None
    '''
    node:dict = {
      "key":None, # {str|int} the dict key or the list index
      "value":None, # {Any} the node value (may be a placeholder)
      "parent":None, # {ConfigTree} the parent node 
    }
    
    # Split the path into segments, handling both dot notation and list indices
    # Regex to match either: [index] or .key
    segments = re.split(r'\.(?![^\[]*\])|\[(\d+)\]', path)
    # Filter out empty strings and None values
    segments = [seg for seg in segments if seg is not None and seg != '']
    
    current = tree
    parent = None
    key = None
    
    # Traverse through each segment
    for segment in segments:
      if isinstance(current, list):
        # Current is a list, segment should be an index
        try:
          index = int(segment)
          # Check if index is valid (non-negative and within range)
          if index < 0 or index >= len(current):
            # Invalid index, return node with None values
            return node
          key = index
        except ValueError:
          # Invalid index format (not a valid integer), return node with None values
          return node
      else:
        # Current is a ConfigTree, segment should be a key
        key = segment
        # Check if key exists in current node
        if key not in current:
          return node
      
      # Save parent before moving to next level
      parent = current
      
      # Move to next level
      if isinstance(current, list):
        # If current is a list, get the item at index (already validated)
        current = current[key]
      else:
        # If current is a ConfigTree, get the value by key (already validated)
        current = current[key]
    
    # Set the node information
    node['key'] = key
    node['value'] = current
    node['parent'] = parent
    
    return node


from blues_lib.type.output.SQLSTDOut import SQLSTDOut
from blues_lib.dao.sql.TableQuerier import TableQuerier

class MatQuerier(TableQuerier):

  _TABLE = 'ap_mat'

  def __init__(self) -> None:
    super().__init__(self._TABLE)

  def exist(self,value:any,field:str='mat_id')->bool:
    fields = [field]
    conditions = [
      {
        'field':field,
        'comparator':'=',
        'value':value,
      },
    ]
    stdout:SQLSTDOut = self.get(fields,conditions)
    return stdout.count>0

  def random(self,fields='*',conditions:list[dict]=None,size:int=1)->SQLSTDOut:
    '''
    @description: Get a random row
    @param {dict} query : the query dict,like:
      [
        {"field":"mat_chan","comparator":"=","value":"${mat_chan}"},
        {"field":"mat_stat","comparator":"=","value":"available","operator":"and"},
      ]
    @param {int} size : the size of the random rows
    @returns {SQLSTDOut}
    '''
    # get the latest
    orders = [{
      'field':'rand()',
      'sort':''
    }]
    # get one row
    pagination = {
      'no':1,
      'size':size
    }
    return self.get(fields,conditions,orders,pagination)

  def latest(self,fields='*',conditions:list[dict]=None,size:int=1)->SQLSTDOut:
    '''
    @description: Get the latest rows
    @param {list[dict]} conditions :
      [
        {"field":"mat_chan","comparator":"=","value":"${mat_chan}"},
        {"field":"mat_stat","comparator":"=","value":"available","operator":"and"},
      ]
    @param {int} size : the size of the latest rows
    @returns {SQLSTDOut}
    '''
    # get the latest
    orders = [{
      'field':'mat_ctime',
      'sort':'desc'
    }]
    # get one row
    pagination = {
      'no':1,
      'size':size
    }
    return self.get(fields,conditions,orders,pagination)



import requests
from .JSDocument import JSDocument
from .JSEvent import JSEvent
from .JSLoader import JSLoader
from .JSScroll import JSScroll
from .JSJQuery import JSJQuery
from .JSPrinter import JSPrinter 
from .JSAlert import JSAlert 

# 混入更多行为模块
class JavaScript(JSDocument,JSEvent,JSLoader,JSScroll,JSJQuery,JSPrinter,JSAlert):
  
  def __init__(self,driver):
    self.driver = driver

  
  def execute(self,script,*args):
    '''
    @description : execute the script string
    @param {str} script : js script
    @param {tuple} args : the rest args
    @returns {any} The return value of the js script
    '''
    return self.driver.execute_script(script,*args)

    
  def execute_async(self,script,*args):
    '''
    @description : Execute js script asynchronously
    @param {str} script : js script
    @param {tuple} args : the rest args
    @returns {str} The return value of the js script's callback func
    '''
    return self.driver.execute_async_script(script,*args)

  def execute_online(self,url):
    '''
    @description : Execute the js script online
    @param {str} url : the js script file path
    @returns {str} : The return value of the js script
    '''
    if not url:
      return None
    response = requests.get(url = url)
    return self.execute(response.text)
  
  def is_var_available(self,var):
    '''
    @description : Determines whether the variable is available
    @param {str} var : variable name, Support attribute concatenation
     - 'navigator'
     - 'navigator.userAgent'
    @returns {bool}
    '''
    var_name = 'window.%s' % var
    script = "var global_var = %s; return !!global_var;" % var_name
    return self.execute(script)



from blues_lib.hook.command.processor.material.MatProc import MatProc
from blues_lib.material.validator.Validator import Validator
from blues_lib.type.output.STDOut import STDOut

class MatValidator(MatProc):
  
  def _calculate(self,request:dict)->STDOut:
    handler = Validator(request)
    return handler.resolve()



import sys,os,re
from typing import Any

from blues_lib.behavior.Bean import Bean

class Text(Bean):

  def _get(self)->str:
    kwargs = self._get_kwargs(['loc_or_elem','parent_loc_or_elem','timeout'])
    return self._browser.element.info.get_text(**kwargs)

  def _set(self):
    selector:str = self._config.get('loc_or_elem')
    parent_selector:str = self._config.get('parent_loc_or_elem')
    value:str = self._config.get('value','')
    if self._to_be_presence():
      return self._browser.script.javascript.text(selector,value,parent_selector)



from blues_lib.type.executor.Executor import Executor

class Command(Executor):

  def __init__(self,context:dict):
    super().__init__()
    self._context = context

  @property
  def context(self):
    return self._context




from .Querier import Querier 
from .EC import EC 
from .Expecter import Expecter 

class Waiter():

  def __init__(self,driver):
    self.querier = Querier(driver)
    self.ec = EC(driver)
    self.expecter = Expecter(driver)



import sys,os,re,datetime

from blues_lib.type.output.STDOut import STDOut
from blues_lib.type.file.File import File
from blues_lib.util.BluesDateTime import BluesDateTime    
from blues_lib.util.BluesImage import BluesImage    
from blues_lib.util.FileDownloader import FileDownloader    

class MatFile(File):
  
  # the material stack's root dir
  MATERIAL_DIR = 'material'
  MATERIAL_LOG_DIR = 'log'

  # the material statck's file
  STACK_FILE_NAME = 'stack.json'

  @classmethod
  def get_download_dir(cls,dirs=[]):
    today = BluesDateTime.get_today()
    subdirs = [cls.MATERIAL_DIR,today,*dirs]
    return cls.get_dir_path(subdirs)

  @classmethod
  def get_download_image(cls,site,file_name,url)->STDOut:
    '''
    Download the image in the body
    Parameter:
      site {str} : the site's name
      file_name {str} : the material's file name
      url {str} : the image's online url
    '''
    if not site or not file_name or not url:
      return STDOut(500,'Failed to download - The parameters site,file_name,url are required',None)

    image_dir = cls.get_download_dir([site])
    result = FileDownloader.download_one(url,image_dir,file_name)
    if result[0]==200:
      # convert type and size
      download_path = result[1]
      # the filename may be changed
      converted_path = BluesImage.convert_type(download_path)
      BluesImage.convert_size(converted_path)
      return STDOut(200,'Managed to download',converted_path)
    else:
      return STDOut(500,f'Failed to download - {result[1]}',url)

  @classmethod
  def get_stack_file(cls):
    return cls.get_file_path(cls.MATERIAL_DIR,cls.STACK_FILE_NAME)
  
  @classmethod
  def get_stack_root(cls):
    return cls.get_dir_path(cls.MATERIAL_DIR)

  @classmethod
  def get_material_root(cls):
    return cls.get_dir_path(cls.MATERIAL_DIR)

  @classmethod
  def get_log_root(cls):
    return cls.get_dir_path(cls.MATERIAL_LOG_DIR)

  @classmethod
  def get_today_log_root(cls):
    today = datetime.datetime.now().strftime("%Y-%m-%d")
    root = cls.get_log_root()
    return os.path.join(root,today)



from blues_lib.hook.FuncHandler import FuncHandler

class BehaviorFuncHandler(FuncHandler):
  
  def execute(self):
    value = self._options.get('value')
    # lambda must return a value
    self._options['value'] = self._func(value) 
